#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# loads in monster & move data from yaml configs, sets the data dictionary to 
#   a variable accessible to other files via from getdata import VAR_NAME
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external
import yaml
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
import pathlib
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

def load_monster_data():
    data_path = pathlib.Path(__file__).parent.parent.parent.joinpath('data/settings')
    monster_data_path = data_path.joinpath('monsters.yaml')
    with monster_data_path.open('r') as file:
       global monster_data
       monster_data = yaml.safe_load(file)
       return monster_data

def load_move_data():
    data_path = pathlib.Path(__file__).parent.parent.parent.joinpath('data/settings')
    move_data_path = data_path.joinpath('moves.yaml')
    with move_data_path.open('r') as file:
        global move_data
        move_data = yaml.safe_load(file)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local
from .monster import boss
from .monster import partner
from ..ui import battlescreen
from ..ui.battlescreen import battlescreen
from . import partnerturn
from . import ai
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

def battle():
    battlescreen.render_monsters(partner, boss)
    while boss.hp > 0 and partner.hp > 0:
        #battlescreen = screen.BattleScreen()
        battlescreen.render_healthbar(partner, boss)
        partnerturn.partner_turn()
        # break if either monster is at or below 0 hp after partner turn
        if partner.hp <= 0 or boss.hp <= 0:
            break
        ai.simulate_turn(attacker=boss, defender=partner)

    if boss.hp <= 0:
        battlescreen.victory(partner, boss)
    elif partner.hp <= 0:
        battlescreen.defeat(partner, boss)
    else:
        raise ValueError('Error: battle ended with neither monster losing '\
                'their full hp.')




#===============================================================================
# aibou wrapper script
#===============================================================================
# order of imports is important
# game data handlers
def play():
    from .battle import getdata
    getdata.load_monster_data()
    getdata.load_move_data()
    from .ui import battlescreen
    from .ui import qtescreen
    from .ui import homescreen
    battlescreen.make_battlescreen()
    qtescreen.make_qtescreen()
#===============================================================================
    from .battle import monster
    monster.create_partner('babybee')
    monster.create_boss('centipede')
    homescreen.start()
    from .battle import runbattle
#===============================================================================
# homescreen
# show home screen
#===============================================================================
# quickplay
# create monsters
# load monsters into namespace
#from monster import partner
#from monster import boss

# make battlescreens

# start battle
    runbattle.battle()



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
import random
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local
from .import battlemechanics
from ..ui.battlescreen import battlescreen # load battlescreen instance into namespace
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

def get_move_data(attacker):
    return attacker.moveset.dict

def initialize_weight_dict():
    global weight_dict
    weight_dict = dict()

def update_weight_dict(attacker, skill_dict_with_move_lists, distribute_weight='select'):
    '''
    Update weight_dict based on moves skill

    Parameters
    -----

    attacker [obj Monster] - The attacking monster

    skill_dict_with_move_lists [dict] - A dictionary with skills and
    moves as values.

    distribute_weight [string] - A string of either 'select' or 'all'

    Key Variables
    -----

    weight_dict [global dict] - Dictionary with moves as keys and weights as
    values. Weight assignment is guided through control flow until values sum
    to 1.

    remainder - One minus the sum of weights already assigned.

    remainder portion - Used to distribute weights. It's value is the
    quotient of the remainder divided by the number of moves that will recieve
    the portion.
    '''
    total_moves_in_skill_dict = 0 # initialize counter
    for move_list in skill_dict_with_move_lists.values():
        total_moves_in_skill_dict += len(move_list)
    if total_moves_in_skill_dict == 0:
        return
    for skill, move_list in skill_dict_with_move_lists.items():
        for move in move_list:
            if skill == 'easy':
                weight_dict[move] += 0.25
            elif skill == 'medium':
                weight_dict[move] += 0.2
            elif skill == 'hard':
                weight_dict[move] += 0.15
            else:
                pass
    sum_weights = sum(weight_dict.values())
    remainder = 1 - sum_weights
    if distribute_weight == 'select':
        remainder_portion = remainder / total_moves_in_skill_dict
        # distribute leftover weights to select moves
        for move,weight in weight_dict.items():
            for skill,move_list in skill_dict_with_move_lists.items():
                if move in move_list:
                    weight_dict[move] += remainder_portion
    elif distribute_weight == 'all':
        # distribute to all moves
        move_data = get_move_data(attacker)
        remainder_portion = remainder / len(move_data.keys())
        for move, data in move_data.items():
            weight_dict[move] += remainder_portion
    print('weight dict: ', weight_dict)
    return

def is_weight_dict_full():
    if sum(weight_dict.values()) > 0.99 and sum(weight_dict.values()) < 1.01:
        print('sum of weight_dict.values', sum(weight_dict.values()), weight_dict)
        print('weight dict is full')
        return True
    else:
        return False

def test_weight_dict_full(func):

    def wrapper(*args, **kwargs):
        if sum(weight_dict.values()) > 0.99 and sum(weight_dict.values()) < 1.01:
            print('sum of weight_dict.values', sum(weight_dict.values()), weight_dict)
            print('weight dict is full')
        else:
            return func(*args, **kwargs)

    return wrapper

def check_kill(attacker, defending_monster):
    ''' Adds weight to moves that can kill defending_monster '''
    move_data = get_move_data(attacker)
    final_blows = {'easy': [], 'medium': [], 'hard': []}
    for move,data in move_data.items():
        weight_dict[move] = 0.0
        if move_data[move]['power'] >= defending_monster.hp:
            skill = data['skill']
            final_blows[skill].append(move)
            print(f'kill move detected: {move}')
    update_weight_dict(attacker, final_blows, distribute_weight='select')

@test_weight_dict_full
def check_heal(attacker):
    ''' Boss attempts to heal or dodge when below 50% hp '''
    #if is_weight_dict_full() == True:
    #    return
    if attacker.hp >= 0.5 * attacker.max_hp:
        print('check heal is not assigning weights')
        return
    else: # add weights to healing and dodging moves
        print('heal priority detected')
        move_data = get_move_data(attacker)
        heal_and_evade_moves = {'easy': [], 'medium': [], 'hard': []}
        for move,data in move_data.items():
            # check if there's a healing move
            try: # lifesteal_portion is not always specified in move config
                if data['lifesteal_portion']:
                    heal_and_evade_moves[data['skill']].append(move)
            except KeyError:
               pass
            if data['type'] == 'evade':
                heal_and_evade_moves[data['skill']].append(move)

        update_weight_dict(
                attacker,
                heal_and_evade_moves,
                distribute_weight='all'
                )
        return

@test_weight_dict_full
def finalize_weight_dict(attacker):
    '''
    Finalizes weight_dict by checking if weights add to 1. In the case of
    a detected kill or detected heal, then the weights should already be 1.
    If not, then this function will distribute the weights uniformly to each move.
    '''
    # if weights are not already one, distribute them uniformly
    if is_weight_dict_full() == False:
        moveset_dict = get_move_data(attacker)
        total_moves = len(moveset_dict.keys())
        for move in moveset_dict.keys():
            weight_dict[move] += (1 / total_moves)
        print('finalizing weight_dict', weight_dict)
    # final check for weights
    if is_weight_dict_full() == False:
        raise ValueError('ai move weights do not add to one', weight_dict)
    return

def choose_weighted_moves():
    options = []
    weights = []
    for move, weight in weight_dict.items():
        options.append(move)
        weights.append(weight)
    selection = random.choices(options, weights=weights, k=1)[0]
    return selection

def simulate_qte(selected_move, attacker):
    moveset_dict = get_move_data(attacker)
    move_data = moveset_dict[selected_move]
    max_damage = move_data['power']
    skill = move_data['skill']
    num_events = move_data['num_events']
    if skill == 'easy':
        p_success = 0.9
    elif skill == 'medium':
        p_success = 0.8
    elif skill == 'hard':
        p_success = 0.7
    else:
        raise ValueError('skill must be easy, medium, or hard')
    p_fail = 1 - p_success
    weights = [p_fail, p_success]
    success_count = 0
    for event in range(num_events):
        outcome = random.choices([0,1], weights=weights, k=1)[0] # get item from single element list
        success_count += outcome
    return success_count

def check_lifesteal(attacker, selected_move):
    move_data = get_move_data(attacker)
    try:
        lifesteal_portion = move_data[selected_move]['lifesteal_portion']
        return lifesteal_portion
    except KeyError:
        return None

def simulate_turn(attacker, defender):
    initialize_weight_dict()
    #while is_weight_dict_full() == False:
    check_kill(attacker, defender)
    check_heal(attacker)
    finalize_weight_dict(attacker)
    selection = choose_weighted_moves()
    print(weight_dict)
    battlescreen.show_move_usage(attacker, selection)
    success_count = simulate_qte(
            selection, attacker
            )
    battlemechanics.resolve_move(attacker, defender, selection, success_count)
    battlescreen.render_healthbar(partner=defender, boss=attacker)
    return



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external 
import yaml
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std 
import os
import pathlib
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local
from . import getdata
from .getdata import monster_data, move_data
from ..ui.color import red, blue, green, yellow
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
class Moveset():
    ''' Store move data for a specific monster '''

    def __init__(self, monster_name):
        self.monster_name = monster_name
        self.move_names = monster_data[monster_name]['moveset']
        self.dict = dict()
        for move in self.move_names:
            self.dict[move] = move_data[move]

class Monster():

    _instances = []

    def __init__(self, name):
        self.name = name
        self.art_file = name + '.txt'
        self.alive = True
        self.status = dict()
        # get art 
        art_path = pathlib.Path(__file__).parent.parent.parent.joinpath('data/monsters')
        art_file_path = art_path.joinpath(self.art_file)
        with art_file_path.open('r') as file:
            lines = []
            for line in file:
                lines.append(line)
        # extra newline leaves space for move result pop-ups
        self.text = '\n' + ''.join(lines)
        self.height = self.text.count('\n')

        # get monster data
        self.data = monster_data[name]
        self.hp = self.data['hp']
        self.max_hp = self.hp
        self.moveset = Moveset(self.name)

        self._instances.append(self)

    def display(self):
        print(self.text)

    def color(self, choice):
        if choice in ['red', 'blue', 'green', 'yellow']:
            self.text = choice(self.text)

    def check_health(self):
        if self.hp > 0:
            pass
        else:
            self.alive = False
    
    def update_hp(self, delta):
        self.hp = self.hp - delta
        check_health()

class Partner(Monster):

    def __init__(self, name):
        Monster.__init__(self, name)
        self.type = 'partner'


class Boss(Monster):

    def __init__(self, name):
        Monster.__init__(self, name)
        self.type = 'boss'

def create_partner(name):
    global partner
    partner = Partner(name)

def create_boss(name):
    global boss
    boss = Boss(name)




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external
import yaml
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
import os
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local
from . import ai
from ..ui import screen
from ..ui.battlescreen import battlescreen
from ..ui.qtescreen import qtescreen
from . import battlemechanics
from .getdata import monster_data, move_data
from . import monster
from .monster import partner
from .monster import boss
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
def deal_damage(monster, power, num_events, num_successes):
    monster.hp = monster.hp - (power * (num_events / num_successes))

def calc_evade_result(base_evasion_stat, num_successes, num_events):
   return base_evasion_stat * (num_successes / num_events)

def resolve_move(selection, type, num_successes):
    move_type = move_data[selection]['type'] 
    power = move_data[selection]['power'] 
    num_events = move_data[selection]['num_events'] 
    if 'damage' in move_type:
        deal_damage(boss, power, num_events, num_successes)
    elif 'evade' in move_type:
        efficacy = move_data[selection]['efficacy']
        calc_evade_result(efficacy, num_successes, num_events)
    elif 'status' in move_type:
        pass
    elif len(move_type == 0):
        raise ValueError(f'The move, {selection}\'s type list is empty.')
    return

def partner_turn():
    battlescreen.render_healthbar(partner, boss)
    move = battlescreen.prompt_move(partner)
    move_type = move_data[move]['type'] 
    num_events = move_data[move]['num_events'] 
    event_time = move_data[move]['event_time'] 

    battlescreen.show_move_usage(partner, move)
    num_successes = qtescreen.start(num_events, event_time)
    battlemechanics.resolve_move(partner, boss, move, num_successes)
    battlescreen.render_healthbar(partner, boss)



import yaml

def set_difficulty(selection):
    # load settings file
    with open('../data/settings.yaml', 'r') as settings_file:
        settings = yaml.safe_load(settings_file)
        difficulty = settings['difficulty']
        # check if current setting is the same as the selection
        if selection == difficulty:
            return
        # update difficulty setting
        settings['difficulty'] = selection
    with open('../data/settings.yaml', 'w') as settings_file:
        # write to file
        yaml.dump(settings, settings_file)




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external packages
import keyboard # module requires root privleges; check bash_aliases for ex on testing
from rich.align import Align
from rich.layout import Layout
from rich.layout import Panel
from rich.padding import Padding
from rich import print
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
import random
from time import sleep
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local modules
from .screen import Screen
from ..battle import qte
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class QTEScreen(Screen):
    
    def __init__(self):
        Screen.__init__(self)

        self.layout.split_column(
                Layout(name='eventspace'),
                Layout(name='lower')
                )
        self.layout['lower'].size = 4

    def start(self, num_events, event_time):
        ''' Setup qte display and call qte functions for updates '''
        result_tally = 0
        for i in range(num_events):
# ============================================================================== 
            self.layout['lower'].update(Panel('Press the key!'))
            sleep(1.5) # delay before event starts
            character = qte.randomcharacter()
            # display char
            self.layout['eventspace'].update(
                    Panel(
                        Align(
                            Padding(character, (1,1), style = 'on blue'),
                            align=random.choice(['left', 'center', 'right']),
                            vertical=random.choice(['top', 'middle', 'bottom'])
                            )
                        )
                    )
            self.show()

            result, feedback = qte.runevent(character, event_time)
            if result == 0:
                self.layout['eventspace'].update(
                        Panel(
                            Align(
                                Padding(feedback, (1,1), style = 'on red'),
                                align='center',
                                vertical='middle')
                            )
                        )
                self.layout['lower'].update(
                        Panel(
                            Align(
                                Padding('Fail', style = 'on red'),
                                align='center',
                                vertical='middle')
                            )
                        )
            elif result == 1:
                self.layout['eventspace'].update(
                        Panel(
                            Align(
                                Padding(feedback, (1,1), style = 'on green'),
                                align='center',
                                vertical='middle')
                            )
                        )
                self.layout['lower'].update(
                        Panel(
                            Align(
                                Padding('Success!', style = 'on green'),
                                align='center',
                                vertical='middle')
                            )
                        )
                result_tally += 1
            else:
                raise ValueError(
                        'Unexpected value for <result>',
                        result
                        )
            # display result
            self.show()
# ============================================================================== 
        sleep(1.5) # pause and show final feedback before moving to damage step
        self.show()
        return result_tally

def make_qtescreen():
    global qtescreen
    qtescreen = QTEScreen()




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external
import rich
import keyboard
from rich.padding import Padding
from rich.align import Align
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
from time import sleep
import random 
import string
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
random.seed=0

class Event():
    def __init__(self):
        # default values are for failure
        self.result = 0
        self.feedback = 'X'
        # updated by success()

def randomcharacter():
    ''' Returns a random a-z or 0-9 character. '''
    return random.choice(string.ascii_lowercase + string.digits)

def success(char):
    print(f'success {char} was pressed')
    keyboard.remove_hotkey(char) # prevent new success calls
    event.result += 1
    event.feedback = u'\u2713'

def fail(char):
    print('fail')
    checkcharacter(char, callback=False)

def addhotkey(char):
    keyboard.add_hotkey(char, success, args=[char])

def removehotkey(char):
    keyboard.remove_hotkey(char)

def checkcharacter(char, time):
    global event
    event = Event()
    addhotkey(char)
    sleep(time)
    # handle removing key only when success() wasn't called (failed qte)
    try:
        removehotkey(char)
    except KeyError: 
        pass
    return event.result, event.feedback

def runevent(char, time):
    result, feedback = checkcharacter(char, time)
    return result, feedback

    




# color.py
# ============================================================================== 
from termcolor import colored

def red(text):
    return colored(text, 'red')

def green(text):
    return colored(text, 'green')

def blue(text):
    return colored(text, 'blue')

def yellow(text):
    return colored(text, 'yellow')



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external
import keyboard
from rich.align import Align
from rich.layout import Layout
from rich.text import Text
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
import pathlib
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local
from .screen import Screen
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\


class HomeScreen(Screen):

    def __init__(self):
        Screen.__init__(self)

#def render_monsters(self, partner, boss):
#    # fit monsters in layout height
#    for monster in [partner, boss]:
#        self.fit_monster(monster)
#    self.partner_render = Align(partner.text, align='left', vertical='bottom')
#    self.boss_render = Align(boss.text, align='right', vertical='top')
#    columns = Columns([self.partner_render, self.boss_render], expand=True)
#    self.layout['middle'].update(Panel(columns))
#    return


def load_art(filename):
    art_path = pathlib.Path(__file__).parent.parent.parent.joinpath('data/ui_art')
    art_path = art_path.joinpath(filename)
    with art_path.open('r') as file:
        lines = []
        for line in file:
            lines.append(line)
        art = ''.join(lines)
    return art

def welcome_text():
    description = Text('Welcome to Aibou, a turn-based monster battle game.\n' \
            "For optimal experience, expand terminal to max window size.")
    return description

def info_option(key):
    keyboard.remove_hotkey(key)
    # open doc or show text desc

def story_option(key):
    keyboard.remove_hotkey(key)
    # enter story mode

def quickplay_option(key):
    keyboard.remove_hotkey(key)
    # enter quick play

def config_option(key):
    keyboard.remove_hotkey(key)
    # enter quick play

def set_menu_option(layout_name, option):
    homescreen.layout[layout_name].update(
        Align(
            # babyblue: https://www.canva.com/colors/color-palettes/mermaid-lagoon/
            Text(option, '#B1D4E0'),
            align='center',
            vertical='middle'
            )
    )

def set_menu_ui(title_art, description):
    homescreen.layout.split_column(
        Layout(name='title', ratio=30),
        Layout(name='options', ratio=65),
        Layout(name='creator', ratio=5)
    )
    homescreen.layout['title'].update(
            Align(title_art + '\n' + description,
                        align='center',
                        vertical='middle'
                 )
             )
    homescreen.layout['options'].split_column(
        Layout(name='quickplay'),
        Layout(name='story'),
        Layout(name='config'),
        Layout(name='info')
    )
    set_menu_option('quickplay', load_art('quickplay.ascii'))
    set_menu_option('story', load_art('story.ascii'))
    set_menu_option('config', load_art('config.ascii'))
    set_menu_option('info', load_art('info.ascii'))
    #keyboard.add_hotkey('q', args='q', callback=quickplay_option)
    keyboard.add_hotkey('i', args='i', callback=info_option)
    keyboard.add_hotkey('c', args='c', callback=info_option)
    keyboard.add_hotkey('o', args='o', callback=info_option)
    homescreen.layout['creator'].update(
            Align(Text('Created by Jake Krol (2023)', '#2E8BC0'),
                  align='center', vertical='middle')
            )
def start():
    title_art = Text(load_art('title.ascii'), '#2E8BC0')
    description = welcome_text()
    global homescreen
    homescreen = HomeScreen()
    set_menu_ui(title_art, description)
    homescreen.show()
    #choose_partner()
    keyboard.wait('q')

#def make_welcomescreen():
#    welcomescreen = WelcomeScreen()
#    return welcomescreen


#def choose_partner():
#    art_path = pathlib.Path(__file__).parent.parent.joinpath('monster-art')
#    partners = os.listdir(art_path)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
import os
import random
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external
import yaml
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local
from . import ai
from ..ui.battlescreen import battlescreen
from . import monster
from .getdata import monster_data, move_data
from .monster import partner
from .monster import boss
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
def deal_damage(monster, power, num_events, num_successes):
    damage = power * (num_successes / num_events)
    monster.hp = monster.hp - damage
    return damage

def resolve_heal(monster, power, num_events, num_successes):
    heal_amount = power * (num_successes / num_events)
    if (monster.hp + heal_amount) > monster.max_hp:
        # recalculate heal amount to be 
        heal_amount = monster.max_hp - monster.hp
    monster.hp += heal_amount
    return heal_amount

def resolve_lifesteal(attacker, selected_move, damage):
    lifesteal_portion = move_data[selected_move]['lifesteal_portion']
    attacker.hp += lifesteal_portion * damage
    # prevent overhealing
    if attacker.hp > attacker.max_hp:
        attacker.hp = attacker.max_hp
    return lifesteal_portion

def calc_evade_result(monster, base_evasion_stat, num_successes, num_events):
    evasion_success_chance = base_evasion_stat * (num_successes / num_events)
    weights = [evasion_success_chance, 1 - evasion_success_chance]
    result = random.choices(['success', 'fail'], weights, k=1)[0]
    return result

def resolve_evade(defender):
    efficacy = defender.status['evading']['efficacy']
    result = random.choices(['success', 'fail'], [efficacy, 1 - efficacy], k=1)[0]
    defender.status['evading']['duration'] -= 1
    duration = defender.status['evading']['duration']
    if defender.status['evading']['duration'] == 0:
    # remove evading status with dictionary comprehension filter
        defender.status = \
                {status:info for (status,info) in defender.status.items() if status != 'evading'}

    return result, duration

def resolve_move(attacker, defender, selection, num_successes):
    move_type = move_data[selection]['type'] 
    power = move_data[selection]['power'] 
    num_events = move_data[selection]['num_events'] 
    # apply move and print outcome to screen
    if 'damage' in move_type:
        if 'evading' in defender.status.keys():
            result, duration = resolve_evade(defender)
            battlescreen.show_evade_outcome(attacker, defender, selection, result, duration)
            if result == 'success':
                return 
        damage = deal_damage(defender, power, num_events, num_successes)
        outcome_text = f'{attacker.name} dealt {damage} damage to {defender.name}!'
        target = defender.type
        battlescreen.show_damage(partner, boss, target, damage)
        if 'lifesteal' in move_type:
            lifesteal_portion = resolve_lifesteal(attacker, selection, damage)
            outcome_text = f'{attacker.name} dealt {damage} damage to {defender.name} '\
            f'and healed {lifesteal_portion * power}!'
    elif 'heal' in move_type:
        heal_amount = resolve_heal(attacker, power, num_events, num_successes)
        target = attacker.type
        battlescreen.show_heal(partner, boss, target, heal_amount)
        outcome_text = f'{attacker.name} healed {heal_amount}hp!'
    elif 'status' in move_type:
        if move_data[selection]['effect']['type'] == 'evade':
            efficacy = move_data[selection]['effect']['efficacy']
            duration = move_data[selection]['effect']['duration']
            result = calc_evade_result(attacker, efficacy, num_successes, num_events)
            if result == 'success':
                attacker.status['evading'] = {'efficacy': efficacy, 'duration': duration}
                outcome_text = f'{attacker.name} will attempt to evade the next '\
                        f'{duration} attacks!'
            elif result == 'fail':
                outcome_text = f'{attacker.name}\'s evasion attempt failed!'
    elif len(move_type) == 0:
        raise ValueError(f'The move, {selection}\'s type list is empty.')
    battlescreen.show_move_outcome(
            attacker, defender, selection, num_events, num_successes, outcome_text
            )
    return




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external packages
from rich import print
from rich.layout import Layout
from rich.layout import Panel
from rich.prompt import Prompt
from rich.align import Align
from rich.columns import Columns
from rich.text import Text
import keyboard
from termios import tcflush, TCIFLUSH
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# std
import sys
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# local modules
from ..battle import monster
from .screen import Screen
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\

class Healthbar():

    def __init__(self, monster):
        max_hearts = int(monster.max_hp // 15)
        num_hearts = int(monster.hp // 15)
        num_Xs = max_hearts - num_hearts

        if monster.type == 'partner':
            self.namecolor = 'green'
        elif monster.type == 'boss':
            self.namecolor = 'red'
        else:
            raise ValueError(
                    'Bad value for monster.type during health bar ' \
                    'initializaiton.\nIs a Monster object being used for ' \
                    'battle instead of a Partner or Boss?'
                    )
        # make final heart stay while monster is alive
        if num_hearts == 0 and monster.hp > 0:
            num_hearts = 1
        # prevent negative hp display on death
        if monster.hp <= 0:
            hp_text = '0'
        else:
            hp_text = str(round(monster.hp, 1))
        nametext = Text(monster.name + ': ')
        nametext.stylize('italic ' + self.namecolor + ' on white')
        hearttext = Text(('\u2665' * num_hearts) + ('X' * num_Xs) + ' ' + hp_text)
        hearttext.stylize('red on white')
        # add status text
        if monster.status.keys():
            if 'evading' in monster.status.keys():
                statustext = Text('E')
                statustext.stylize('italic white on purple')
                # partner status on right of healthbar
                if monster.type == 'partner':
                    self.text = nametext + hearttext + ' ' + statustext
                # boss status on left of health bar
                if monster.type == 'boss':
                    self.text = statustext + ' ' + nametext + hearttext
        else:
            self.text = nametext + hearttext


class BattleScreen(Screen):

    def __init__(self):
        Screen.__init__(self)

        self.layout.split_column(
                Layout(name='upper'),
                Layout(name='middle'),
                Layout(name='lower')
                )
        self.layout['middle'].ratio = 4
        self.layout['lower'].size = 4
        self.layout['upper'].size = 3

    def fit_monster(self, monster):
        self.layout['middle'].minimum_size = monster.height + 2 # add 2 lines for panel outline

    def render_monsters(self, partner, boss):
        # fit monsters in layout height
        for monster in [partner, boss]:
            self.fit_monster(monster)
        self.partner_render = Align(partner.text, align='left', vertical='bottom')
        self.boss_render = Align(boss.text, align='right', vertical='top')
        columns = Columns([self.partner_render, self.boss_render], expand=True)
        self.layout['middle'].update(Panel(columns))
        return

    def unrender_monster(self, monster):
        if monster.type == 'partner':
            columns = Columns([' ', self.boss_render], expand=True)
        elif monster.type == 'boss':
            columns = Columns([self.partner_render], ' ', expand=True)
        else:
            raise ValueError('wrong value for Monster.type')
        self.layout['middle'].update(Panel(columns))
        return

    def show_damage(self, partner, boss, target, damage):
        partner_render = Text(partner.text)
        boss_render = Text(boss.text)
        if target == 'partner':
            partner_render.append(f'-{str(damage)}', style='bold red')
        elif target == 'boss':
            boss_render.append(f'-{str(damage)}', style='bold red')
        partner_render = Align(partner_render, align='left', vertical='bottom')
        boss_render = Align(boss_render, align='right', vertical='top')
        columns = Columns([partner_render, boss_render], expand=True)
        self.layout['middle'].update(Panel(columns))
        self.show()
        self.pause(2)
        self.render_monsters(partner, boss)

    def show_heal(self, partner, boss, target, heal_amount):
        partner_render = Text(partner.text)
        boss_render = Text(boss.text)
        if target == 'partner':
            partner_render.append(f'+{str(heal_amount)}', style='bold green')
        elif target == 'boss':
            boss_render.append(f'+{str(heal_amount)}', style='bold green')
        partner_render = Align(partner_render, align='left', vertical='bottom')
        boss_render = Align(boss_render, align='right', vertical='top')
        columns = Columns([partner_render, boss_render], expand=True)
        self.layout['middle'].update(Panel(columns))
        self.show()
        self.pause(2)
        self.render_monsters(partner, boss)

    def render_healthbar(self, partner, boss):
        partner_hp_render = Align(
                Healthbar(partner).text,
                align='left',
                vertical='bottom'
                )
        boss_hp_render = Align(Healthbar(boss).text,
                               align='right',
                               vertical='bottom'
                               )
        columns = Columns([partner_hp_render, boss_hp_render], expand=True)
        self.layout['upper'].update(Panel(columns))
        self.show()

    def show_move_usage(self, attacking_monster, selected_move):
        self.layout['lower'].update(
                Panel(f'{attacking_monster.name} uses {selected_move}')
                )
        self.show()
        self.pause(1.5)

    def prompt_move(self, monster):
        tcflush(sys.stdin, TCIFLUSH)# clear stdin queue to prevent entering old key presses
        move_data = monster.moveset.dict
        # initialize options dict and prompt list
        options = {'0': 'evade'}
        prompts = ['0->evade']
        for index,move in enumerate(monster.moveset.move_names):
            prompts.append(f'{index + 1}->{move}')
            options[str(index + 1)] = move
        prompts = '    '.join(prompts)
        self.layout['lower'].update(Panel(f"Choose an attack:\n{prompts}"))
        self.show()
        def show_move_data(key):
            print(f'\n{options[key]}: {move_data[options[key]]}')
            tcflush(sys.stdin, TCIFLUSH)
        for key in options.keys():
            keyboard.add_hotkey(
                    f'i+{key}', show_move_data, args=[key]
                    )
        selection = Prompt.ask(choices=options.keys())

        if selection in options:
            for key in options.keys():
                keyboard.remove_hotkey(f'i+{key}')
            self.show_move_usage(monster, options[selection])
        else:
            self.prompt_move(monster)
        return options[selection]

    def show_evade_outcome(self, attacker, defender, move, result, duration):
        if result == 'success':
            result_message = f'{defender.name} successfully evaded {attacker.name}\'s '\
                    f'{move}!'
        if result == 'fail':
            result_message = f'{defender.name}\'s evade attempt failed!'
        if duration > 0:
            duration_message = f'{defender.name} will attempt to evade {duration} more '\
                    f'attack(s).'
        if duration == 0:
            duration_message = f'{defender.name} is no longer evading.'
        message = result_message + '\n' + duration_message
        self.layout['lower'].update(Panel(message))
        self.show()
        self.pause(3)

    def show_move_outcome(self, attacker, defender, move, num_events, num_successes, bonus_text):
        message = f'{attacker.name} passed {num_successes}/{num_events} skill '\
                'checks. ' + bonus_text
        self.layout['lower'].update(Panel(message))
        self.show()
        self.pause(3)

    def victory(self, partner, boss):
        self.unrender_monster(boss)
        self.layout['lower'].update(
                Panel(f'{partner.name} defeated {boss.name}!')
        )
        self.show()

    def defeat(self, partner, boss):
        self.unrender_monster(partner)
        self.layout['lower'].update(
                Panel(f'{partner.name} has been defeated.')
        )
        self.show()

def make_battlescreen():
    global battlescreen
    battlescreen = BattleScreen()



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
# external packages
from rich import print
from rich.layout import Layout
from time import sleep
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\
class Screen():
    ''' 
    The parent class for all UIs/Screens. An instance initializes a rich Layout.
    '''

    def __init__(self):
        self.layout = Layout()

    def show(self):
        print(self.layout)

    def pause(self, seconds):
        sleep(seconds)

class Menu(Screen):
    pass

